<%- include('includes/menu-item', {
  url: '/foo'
, title: 'Foo'
}); -%>

<%- include('includes/menu-item', {
  url: '/bar'
, title: 'Bar'
}); -%>

<%- include('includes/menu-item', {
  url: '/baz'
, title: 'Baz'
}); -%>

