<% if (names.length) { %>
  <ul>
    <% names.forEach(function (name) { %>
      <%# Notice how the single quotation mark is escaped in the output HTML %>
      <li foo='<%= name + "'" %>'><%= name %></li>
    <% }) %>
  </ul>
<% } %>
