   <tag1>  
<tag2>  
A very long piece of text very long piece of text very long piece of
text very long piece <% var f = 'f' %>of text very long piece of
tex
t very long piece of<% %>text very long 
adsffadsfadsfad<%= f %>   
   
piece of text.
<% var a = 'a' %>  
Text again.
<% var b = 'b' %>
<% var c = 'c'
var d = 'd' %>
Another text. <%= c %>  
