<% var url = '/foo' -%>
<% var title = 'Foo' -%>
<% include includes/menu-item -%>

<% var url = '/bar' -%>
<% var title = 'Bar' -%>
<% include includes/menu-item -%>

<% var url = '/baz' -%>
<% var title = 'Baz' -%>
<% include includes/menu-item -%>
